@@ -463,10 +463,12 @@ static gboolean wispr_route_request(const char *address, int ai_family,
 		break;
 	}
 
-	if (result < 0)
+	if (result < 0) {
+		g_free(route);
 		return FALSE;
+	}
 
-	route->address = strdup(address);
+	route->address = g_strdup(address);
 	route->if_index = if_index;
 	wp_context->route_list = g_slist_prepend(wp_context->route_list, route);
 
@@ -518,17 +520,46 @@ static gboolean wispr_input(const guint8 **data, gsize *length,
 	return FALSE;
 }
 
+static void wispr_portal_browser_reply_cb(struct connman_service *service,
+					connman_bool_t authentication_done,
+					const char *error, void *user_data)
+{
+	struct connman_wispr_portal_context *wp_context = user_data;
+
+	DBG("");
+
+	if (service == NULL || wp_context == NULL)
+		return;
+
+	if (authentication_done == FALSE) {
+		wispr_portal_error(wp_context);
+		free_wispr_routes(wp_context);
+		return;
+	}
+
+	/* Restarting the test */
+	__connman_wispr_start(service, wp_context->type);
+}
+
 static void wispr_portal_request_wispr_login(struct connman_service *service,
 				connman_bool_t success,
 				const char *ssid, int ssid_len,
 				const char *username, const char *password,
 				gboolean wps, const char *wpspin,
-				void *user_data)
+				const char *error, void *user_data)
 {
 	struct connman_wispr_portal_context *wp_context = user_data;
 
 	DBG("");
 
+	if (error != NULL && g_strcmp0(error,
+			"net.connman.Agent.Error.LaunchBrowser") == 0) {
+		__connman_agent_request_browser(service,
+				wispr_portal_browser_reply_cb,
+				wp_context->redirect_url, wp_context);
+		return;
+	}
+
 	g_free(wp_context->wispr_username);
 	wp_context->wispr_username = g_strdup(username);
 
@@ -611,27 +642,6 @@ static gboolean wispr_manage_message(GWebResult *result,
 	return FALSE;
 }
 
-static void wispr_portal_browser_reply_cb(struct connman_service *service,
-					connman_bool_t authentication_done,
-					void *user_data)
-{
-	struct connman_wispr_portal_context *wp_context = user_data;
-
-	DBG("");
-
-	if (service == NULL || wp_context == NULL)
-		return;
-
-	if (authentication_done == FALSE) {
-		wispr_portal_error(wp_context);
-		free_wispr_routes(wp_context);
-		return;
-	}
-
-	/* Restarting the test */
-	__connman_wispr_start(service, wp_context->type);
-}
-
 static gboolean wispr_portal_web_result(GWebResult *result, gpointer user_data)
 {
 	struct connman_wispr_portal_context *wp_context = user_data;
@@ -699,6 +709,7 @@ static gboolean wispr_portal_web_result(GWebResult *result, gpointer user_data)
 				wispr_route_request, wp_context);
 
 		goto done;
+	case 400:
 	case 404:
 		if (__connman_service_online_check_failed(wp_context->service,
 							wp_context->type) == 0)
@@ -724,9 +735,6 @@ static void proxy_callback(const char *proxy, void *user_data)
 
 	wp_context->token = 0;
 
-	if (getenv("CONNMAN_WEB_DEBUG"))
-		g_web_set_debug(wp_context->web, web_debug, "WEB");
-
 	if (proxy != NULL && g_strcmp0(proxy, "DIRECT") != 0)
 		g_web_set_proxy(wp_context->web, proxy);
 
@@ -791,22 +799,28 @@ static int wispr_portal_detect(struct connman_wispr_portal_context *wp_context)
 
 	if_index = connman_inet_ifindex(interface);
 	if (if_index < 0) {
+		DBG("Could not get ifindex");
 		err = -EINVAL;
 		goto done;
 	}
 
 	nameservers = connman_service_get_nameservers(wp_context->service);
 	if (nameservers == NULL) {
+		DBG("Could not get nameservers");
 		err = -EINVAL;
 		goto done;
 	}
 
 	wp_context->web = g_web_new(if_index);
 	if (wp_context->web == NULL) {
+		DBG("Could not set up GWeb");
 		err = -ENOMEM;
 		goto done;
 	}
 
+	if (getenv("CONNMAN_WEB_DEBUG"))
+		g_web_set_debug(wp_context->web, web_debug, "WEB");
+
 	if (wp_context->type == CONNMAN_IPCONFIG_TYPE_IPV4) {
 		g_web_set_address_family(wp_context->web, AF_INET);
 		wp_context->status_url = STATUS_URL_IPV4;
